/*----------------------------------------------------------------------------------
*
* COPYRIGHT (c) 2001 by Singing Electrons, Inc. All rights reserved.
*
*
* Module Name			: C:\se\adi\hidclass\Source\usbdriver.c
*
* Description			: The USB Driver - contains implementation of the low level USB interface.
*
*
* Revision History	: At bottom of the file.
*
*---------------------------------------------------------------------------------*/

/** include files **/
#include <sys/exception.h>
#include <sys/excause.h>
#include <defBF535.h>
#include "dtype.h"
#include "register.h"
#include "usb.h"
#include "usbdriver.h"
#include "dprintf.h"

/** local definitions **/
typedef struct tagDMAALLOC{
	UINT offset;
	UINT size;
}DMAALLOC, *PDMAALLOC;


/* default settings */
#define DMA_BUFFER_SIZE 2048
#define MAX_DMA_BUFFER_ALLOCATIONS 64


typedef struct tagEPSTATE{
	UCHAR currentState;
	USB_SETUP_DATA setupData;
	void *dmaBuffer;
	UINT dmaBufferSize;
	UCHAR endpointType;
	UINT maxPacketSize;
	UINT defaultMaxTransferSize;
} EPSTATE, *PEPSTATE;

//for currentState
enum	{
	EP_STATE_IDLE = 0,
	EP_STATE_WAITING_FOR_SETUP, 
	EP_STATE_DATA_IN_PHASE,
	EP_STATE_DATA_OUT_PHASE,
	EP_STATE_STATUS_PHASE,
	EP_STATE_WAITING_FOR_TC
	};


/** external functions **/

/** external data **/

/** internal functions **/

/** public data **/
#pragma align 128
section("data1") UCHAR gDMABuffer[DMA_BUFFER_SIZE]; 


UINT gTotalDMABuffersAllocated = 0;
DMAALLOC gDMAAllocations[MAX_DMA_BUFFER_ALLOCATIONS];
UINT gLargestFreeDMABlock, gLargestFreeDMABlockOffset;

/** private data **/
static int gNumEPBufsSetup = 0;
static EP_BUF gEndpointBuffers[MAX_EP_BUF];

static EPSTATE gActiveEndpointStates[MAX_PHY_EP];

static PSETUPPROC gpOnSetupReq;

static PGETDEVICEDESCRIPTORPROC gpOnGetDeviceDescriptor;
static PGETCONFIGDESCRIPTORPROC gpOnGetConfigDescriptor;
static PGETSTRINGDESCRIPTORPROC gpOnGetStringDescriptor;

static PTCPROC gpOnTC;
static PMERRPROC gpOnMERR;
static PSUSPENDPROC gpOnSuspend;
static PRESUMEPROC gpOnResume;
static PRESETPROC gpOnReset;
static PCONFIGPROC gpOnConfigurationChange;

static USHORT gep_int_mask; //Nothing special


UCHAR *gpControlEPDMABufferOUT;
UCHAR *gpControlEPDMABufferIN;

/** public functions **/
void *AllocDMABlock (int numBytesToAllocate);
void FreeDMABlock (void *blockToDealloc);
void InitDMA (void);

/** private functions **/
EX_INTERRUPT_HANDLER (USBDriver_ISR);
bool GetControlEndpointReady (UCHAR endpointNumber);
void HandleNoDataControlRequest (UCHAR currentEndpoint);
void HandleDataOutControlRequest (UCHAR currentEndpoint);
void HandleDataInControlRequest (UCHAR currentEndpoint);
bool DispatchDataInControlRequest (UCHAR endpointNumber);


/*------------------------------------------------------------
*	USBDriver_PreInit
*
*	Parameters:  
*
*	Globals Used:
*		gNumEPBufsSetup
*
*	Description:
*		This routine initializes the global state variables used
*		by the driver, and the USBD hardware. It does not
*		initialize the EPBUFs
*
*	Returns:
*
*
*------------------------------------------------------------*/
bool USBDriver_PreInit (void)
	{
		UINT chipVer;


		gNumEPBufsSetup = 0;
		memset (&gEndpointBuffers, 0, sizeof (gEndpointBuffers));

		memset (&gActiveEndpointStates, 0, sizeof (gActiveEndpointStates));

		InitDMA ();

		//Let's setup the control endpoint, as it should be the same
		//for all devices
		USBDriver_SetupEndpoint (EP0, EP_CONFIG0, EP_IFACE0, EP_ALT0, 0, EP_CTL, 
			CONTROL_PIPE_PACKET_SIZE);

		gpOnSetupReq = nil;
		gpOnGetDeviceDescriptor = nil;
		gpOnGetConfigDescriptor = nil;
		gpOnGetStringDescriptor = nil;
		gpOnTC = nil;
		gpOnMERR = nil;
		gpOnSuspend = nil;
		gpOnResume = nil;
		gpOnReset = nil;
		gpOnConfigurationChange = nil;

		// setup endpint interrpt mask
		gep_int_mask = (USBD_GINTR_EP1INT|USBD_GINTR_EP2INT|USBD_GINTR_EP3INT|
					   USBD_GINTR_EP4INT|USBD_GINTR_EP5INT|USBD_GINTR_EP6INT|
					   USBD_GINTR_EP7INT);

		// install the handler 
		register_handler(ik_ivg7, USBDriver_ISR);


		chipVer = 0xf0000000 & chipId_REG;
  
	  
	  if(chipVer) 
	  	{
	    	// rev. 1.0 or higher
		    SIC_IMASK_REG |= SIC_MASK2;
		  } 
		else 
			{
	    	// rev. 0.x
	    	SIC_IMASK_REG &= SIC_UNMASK2;  
	   	}

	  asm("ssync;");

	  IMASK_REG |= 0x0080; // unmask IVG 7 at core
	  asm("ssync;");



		return (true);
	}



/*------------------------------------------------------------
*	USBDriver_SetupEndpoint
*
*	Parameters:
*		endpointNumber - the endpoint number to be used (EP0..EP7)
*		config - which configuration this setting is to be used for.
*		interface - which interface this setting is to be used for.
*		altSetting - which alt. setting this setting is to be used for.
*		direction - could be EP_IN or EP_OUT
*		ep_type - could be EP_CTL, EP_ISO, EP_BULK, or EP_INTERRUPT
*		maxPacketSize - maximum packet length - could be 8, 16, 32, 
*			64 for Bulk, Int, and Control EPs, or any value from
*			0 to 1023 for Iso. endpoints
*
*	Globals Used:
*		gNumEPBufsSetup
*
*	Description:
*		This routine loads the next EP_BUF array member with the 
*		supplied parameters
*
*	Returns:
*		true on success, false on failure
*
*
*------------------------------------------------------------*/
bool USBDriver_SetupEndpoint (UINT endpointNumber, UCHAR altSetting, 
	UCHAR interface, UCHAR configuration, UCHAR direction, 
	UCHAR ep_type, UINT maxPacketSize)			
	{
		//Make sure we can hold them all!
		if (gNumEPBufsSetup >= MAX_EP_BUF)
			return (false);
			
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_bufadrptr = endpointNumber;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_maxpktsize = maxPacketSize;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_dir = direction;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_type = ep_type;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_altsetting = altSetting;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_interface = interface;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.ep_config = configuration;
		gEndpointBuffers[gNumEPBufsSetup].s_ep_buf.epnum = endpointNumber;

		gNumEPBufsSetup++;

		return (true);
		

	}


/*------------------------------------------------------------
*	USBDriver_CompleteInit
*
*	Parameters:  
*
*	Globals Used:
*
*	Description:
*		This routine finishes up the initialization, by installing the
*		callback addresses and loading up the EP bufs, enabling USB, etc..
*
*	Returns:
*
*
*------------------------------------------------------------*/
bool USBDriver_CompleteInit (PSETUPPROC pSetupProc, PGETDEVICEDESCRIPTORPROC pDevDescProc,
	PGETCONFIGDESCRIPTORPROC pCfgDescProc, PGETSTRINGDESCRIPTORPROC pStringDescProc,
	PTCPROC pTCProc, PMERRPROC pMerrProc, PSUSPENDPROC pSuspendProc, PRESUMEPROC pResumeProc,
	PRESETPROC pResetProc, PCONFIGPROC pConfigProc, UINT controlEPMaxTransferSize)
	{
		int epBufNumber, epByte, timeout;

		gpOnSetupReq = pSetupProc;
		gpOnGetDeviceDescriptor = pDevDescProc;
		gpOnGetConfigDescriptor = pCfgDescProc;
		gpOnGetStringDescriptor = pStringDescProc;
		gpOnTC = pTCProc;
		gpOnMERR = pMerrProc;
		gpOnSuspend = pSuspendProc;
		gpOnResume = pResumeProc;
		gpOnReset = pResetProc;
		gpOnConfigurationChange = pConfigProc;

		// initialize all 64 EP through the USBD_EPBUF
		for (epBufNumber = 0; epBufNumber < MAX_EP_BUF; epBufNumber++)
			{
				for (epByte = EP_BUF_SZ; epByte > 0; epByte--)
					{
						UCHAR byteToDownload = gEndpointBuffers[epBufNumber].a_ep_buf[epByte-1];
						USBD_EPBUF_REG = (USHORT) byteToDownload;

						timeout = NUM_EPBUF_RDY_CHK;
						while	(timeout)
							{
								if((USBD_EPBUF_REG) & USBD_EPBUF_RDY)
									break;
								timeout--;
							}

						if (!timeout)
							break;
						asm("ssync;");
					}
			}

		//See if we're all done!
		if(((USBD_EPBUF_REG) & USBD_EPBUF_CFG))
			{
				asm("ssync;");
				return (false);
			}

		asm("ssync;");

		//=========================================================================
		// setup the USBD_GINTR, USBD_INTRx and DMA USBD_DMACFG
		// interrupt mask !
		//=========================================================================
		// clear all pending interrupts
		USBD_GINTR_REG = (USHORT)0xFFFF; 
		
		// unmask all the EPs interrupt as well 
		USBD_GMASK_REG &= (USHORT)~(USBD_GMASK_EP0MSK|USBD_GMASK_EP1MSK|USBD_GMASK_EP2MSK|
									USBD_GMASK_EP3MSK|USBD_GMASK_EP4MSK|USBD_GMASK_EP5MSK|
									USBD_GMASK_EP6MSK|USBD_GMASK_EP7MSK|USBD_GMASK_CFGM);

									
		// clear interrupt for all the EPs
		USBD_INTR0_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT|
								   USBD_INTR_SETUP|USBD_INTR_MSETUP|USBD_INTR_MERR);
		USBD_INTR1_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR2_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR3_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR4_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR5_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR6_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);
		USBD_INTR7_REG |= (USHORT)(USBD_INTR_TC|USBD_INTR_PC|USBD_INTR_BCSTAT);

		// Unmask the USBD interrupts
		USBD_MASK0_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_MASK_SETUP|
							USBD_MASK_MSETUP|USBD_MASK_MERR|USBD_MASK_BCSTAT); 

		// Let's try something different: enable only BCSTAT
		// SL	:)
	#if 1
		USBD_MASK1_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT|USBD_MASK_MERR);
		USBD_MASK2_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
		USBD_MASK3_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
		USBD_MASK4_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
		USBD_MASK5_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
		USBD_MASK6_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
		USBD_MASK7_REG &= (USHORT)~(USBD_MASK_TC|USBD_MASK_PC|USBD_INTR_BCSTAT);
	#else
		USBD_MASK1_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK2_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK3_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK4_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK5_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK6_REG &= (USHORT)~(USBD_INTR_BCSTAT);
		USBD_MASK7_REG &= (USHORT)~(USBD_INTR_BCSTAT);
	#endif
		
		// clear the DMA FIFO first. before we use it !
		USBD_DMACFG_REG |= (USHORT)USBD_DMACFG_DMABC;	
		USBD_DMACFG_REG = (USHORT)USBD_DMACFG_DMAEN;

	 	//=========================================================================
		// program the DMA master interface with the USB module's base address
		// USBD_DMABL, USBD_DMABH
		//=========================================================================
		// BUG ! BUG ! BUG !
		// need to do the lower 16 bit first to avoid a compiler bug !
		USBD_DMABL_REG = (USHORT)((ULONG)gDMABuffer&USBD_DMABL_SGMSK);
		USBD_DMABH_REG = (USHORT)((ULONG)gDMABuffer >> 16);

		ActivateEndpoint (EP0, EP_CTL, CONTROL_PIPE_PACKET_SIZE, controlEPMaxTransferSize);

		if (!GetControlEndpointReady (EP0))
			return (false);


		// Enable the USBD now
		USBD_CTRL_REG |= (USHORT)USBD_CTRL_ENA;

		asm("ssync;");

		//we're now ready to receive interrupts!
		return(true);

	}



/*------------------------------------------------------------
*	ActivateEndpoint
*
*	Parameters:
*		endpointNumber - endpoint to set the information for.
*		endpointType - type of endpoint - EP_CTL, EP_BULK, EP_INT, or EP_ISO
*		maxPacketSize - the maximum packet size for this endpoint.
*		defaultMaxTransferSize - the max. trasfer size to use by default.
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		This routine initializes the active endpoint state for the specified endpoint
*		with the type, max packet size and default max. transfer size for the endpoint.
*		This information is used in other circumstances. This routine should be called
*		once for each endpoint after a new config. has come in effect.
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void ActivateEndpoint (UCHAR endpointNumber, UCHAR endpointType, UCHAR maxPacketSize, UINT defaultMaxTransferSize)
	{
		gActiveEndpointStates[endpointNumber].endpointType = endpointType;
		gActiveEndpointStates[endpointNumber].maxPacketSize = maxPacketSize;
		gActiveEndpointStates[endpointNumber].defaultMaxTransferSize = defaultMaxTransferSize;
	}



/*------------------------------------------------------------
*	ArmEndpoint
*
*	Parameters:  
*		endpointNumber - endpoint to arm.
*		direction - direction for the transfer (EP_IN or EP_OUT)
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		Sets the DMA address/length for an endpoint, and arms it, alongwith providing
*		the appropriate config. information for the endpoint.
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void ArmEndpoint (UCHAR endpointNumber, UCHAR direction)
	{
		USHORT cfgReg;
		//USHORT len = gActiveEndpointStates[endpointNumber].dmaBufferSize;

		cfgReg = (USHORT)((USBD_EPCFG_ARM) | (direction << 1) | (gActiveEndpointStates[endpointNumber].endpointType << 2));
		SET_USBD_MAX(cfgReg, GetMaxPacketSizeCode (gActiveEndpointStates[endpointNumber].maxPacketSize));

		(*(PVUS)((UINT)USBD_EPCFG0 + (USBD_EPCFG1 - USBD_EPCFG0) * endpointNumber)) = cfgReg;
		//Dprintf ("CFG Reg = %08x", cfgReg);

		(*(PVUS)((UINT)USBD_EPADR0 + (USBD_EPCFG1 - USBD_EPCFG0) * endpointNumber)) = 
			(USHORT)((ULONG)gActiveEndpointStates[endpointNumber].dmaBuffer & USBD_OFFSET_SGMSK);
		//Dprintf ("ADR Reg = %08x", (USHORT)((ULONG)gActiveEndpointStates[endpointNumber].dmaBuffer & USBD_OFFSET_SGMSK));

		(*(PVUS)((UINT)USBD_EPLEN0 + (USBD_EPCFG1 - USBD_EPCFG0) * endpointNumber)) = 
			(USHORT)gActiveEndpointStates[endpointNumber].dmaBufferSize;

		//Dprintf ("LEN Reg: %08x", (USHORT)gActiveEndpointStates[endpointNumber].dmaBufferSize);

		asm("ssync;");

	}


/*------------------------------------------------------------
*	USBDriver_ISR
*
*	Parameters:  
*		r0x, r1x, r2x
*
*	Globals Used:
*		
*
*	Description:
*		The ISR to handle USBD interrupts. 
*
*	Returns:
*		Nothing
*
*
*------------------------------------------------------------*/
EX_INTERRUPT_HANDLER  (USBDriver_ISR)
	{
		int	i, currentEndpoint;
		USHORT		usbd_gintr;
		USHORT		usbd_status;
		static USHORT    usbd_epintr[MAX_PHY_EP];
		bool shouldStall;

		// clear the interrupt first
		usbd_gintr = (USHORT)USBD_GINTR_REG;
		USBD_GINTR_REG = (USHORT)usbd_gintr;

		usbd_epintr[0]= (USHORT)USBD_INTR0_REG;
		USBD_INTR0_REG = usbd_epintr[0];
		usbd_epintr[1]= (USHORT)USBD_INTR1_REG;
		USBD_INTR1_REG = usbd_epintr[1];
		usbd_epintr[2]= (USHORT)USBD_INTR2_REG;
		USBD_INTR2_REG = usbd_epintr[2];
		usbd_epintr[3]= (USHORT)USBD_INTR3_REG;
		USBD_INTR3_REG = usbd_epintr[3];
		usbd_epintr[4]= (USHORT)USBD_INTR4_REG;
		USBD_INTR4_REG = usbd_epintr[4];
		usbd_epintr[5]= (USHORT)USBD_INTR5_REG;
		USBD_INTR5_REG = usbd_epintr[5];
		usbd_epintr[6]= (USHORT)USBD_INTR6_REG;
		USBD_INTR6_REG = usbd_epintr[6];
		usbd_epintr[7]= (USHORT)USBD_INTR7_REG;
		USBD_INTR7_REG = usbd_epintr[7];

		usbd_status = (USHORT)USBD_STAT_REG;
		asm("ssync;");

		Dprintf1 ("usbd_gintr = %08x", usbd_gintr);
		Dprintf1 ("epint0 = %08x", usbd_epintr[0]);

		if (usbd_gintr & USBD_GINTR_CFG)
			{
				if (gpOnConfigurationChange)
					{
						//Call the client to do whatever it needs to do on config change
						(*gpOnConfigurationChange)(GET_USBD_CONFIG(usbd_status), 
							GET_USBD_INTERFACE(usbd_status), 
							GET_USBD_ALTSETTING(usbd_status));
					}
			}

		if (usbd_gintr & USBD_GINTR_SUSP)
			{
				//We should reset the states of all our endpoints
				for (currentEndpoint = (MAX_PHY_EP-1); currentEndpoint >= EP0 ; currentEndpoint--)
					{
						gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
						gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;
						gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_IDLE;
					}

				//and free the DMA buffers
				InitDMA ();

				//let the client know about this!
				if (gpOnSuspend)
					(gpOnSuspend)();
			}


		if (usbd_gintr & USBD_GINTR_RESUME)
			{
				//Prepare control EP for receiving a setup packet!
				GetControlEndpointReady (EP0);

				//let the client know about this!
				if (gpOnResume)
					(gpOnResume)();
			}

		if (usbd_gintr & USBD_GINTR_RST)
			{
				//We should reset the states of all our endpoints
				for (currentEndpoint = (MAX_PHY_EP-1); currentEndpoint >= EP0 ; currentEndpoint--)
					{
						gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
						gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;
						gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_IDLE;
					}

				//and free the DMA buffers
				InitDMA ();

				//Prepare control EP for receiving a setup packet!
				GetControlEndpointReady (EP0);
				if (gpOnReset)
					(gpOnReset)();
			}


		for (currentEndpoint = EP0; currentEndpoint < MAX_PHY_EP; currentEndpoint++)
			{
				if ((gActiveEndpointStates[currentEndpoint].currentState == EP_STATE_WAITING_FOR_SETUP) &&
					((usbd_epintr[currentEndpoint] & USBD_INTR_SETUP)))
					{
						//Save a copy of the setup request
						CopyDMABufferToSETUPStruct (&gActiveEndpointStates[currentEndpoint].setupData,
							gActiveEndpointStates[currentEndpoint].dmaBuffer);

						Dprintf ("Received Setup Request");

						//Don't need this buffer any more
						FreeDMABlock (gActiveEndpointStates[currentEndpoint].dmaBuffer);
						gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
						gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;

						//If there's no data-phase to this
						if (gActiveEndpointStates[currentEndpoint].setupData.wLength == 0)
							{
								Dprintf ("Entering Status Phase");
								HandleNoDataControlRequest (currentEndpoint);
							}
						else
							{
								//We come here when wLength is non-zero
								//If direction of accompanying data is OUT, we need to have a DMA buffer
								//to recive it, and ARM our endpoint accordingly
								if (USB_CONTROL_REQUEST_GET_DIRECTION (gActiveEndpointStates[currentEndpoint].setupData.bmRequestType) 
									== USB_CONTROL_REQUEST_TYPE_OUT)
									{
										Dprintf ("Entering Data Out Phase");
										HandleDataOutControlRequest (currentEndpoint);
									}
								else
									{
										Dprintf ("Entering Data In Phase");
										HandleDataInControlRequest (currentEndpoint);
									}
							}
					}
				else if (usbd_epintr[currentEndpoint] & USBD_INTR_TC)
					{
						switch (gActiveEndpointStates[currentEndpoint].currentState)
							{
								case EP_STATE_WAITING_FOR_TC:
									//Free the DMA buffer
									FreeDMABlock (gActiveEndpointStates[currentEndpoint].dmaBuffer);
									gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
									gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;
									gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_IDLE;

									//Let the client know that a xfer has completed
									if (gpOnTC)
										(gpOnTC)(currentEndpoint);
									
									break;
								case EP_STATE_STATUS_PHASE:
									Dprintf ("Completed Status Phase");
									//Nothing more to do, let's get the endpoint ready for more requests
									GetControlEndpointReady (currentEndpoint);
									break;
								case EP_STATE_DATA_OUT_PHASE:
									//We've recevied all the data, let's provide this to client
									Dprintf ("Entering Status Phase from Data Out");
									shouldStall = true;
									if (gpOnSetupReq)
										{
											shouldStall = (!(gpOnSetupReq) (currentEndpoint, &gActiveEndpointStates[currentEndpoint].setupData,
												gActiveEndpointStates[currentEndpoint].dmaBuffer, &gActiveEndpointStates[currentEndpoint].dmaBufferSize));
										}

									//Free the DMA buffer
									FreeDMABlock (gActiveEndpointStates[currentEndpoint].dmaBuffer);
									gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
									gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;

									if (shouldStall)
										{
											StallEndpoint (currentEndpoint);
											GetControlEndpointReady (currentEndpoint);
										}
									else
										{
											gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_STATUS_PHASE;
											ArmEndpoint (currentEndpoint, EP_IN);
										}

									break;
								case EP_STATE_DATA_IN_PHASE:
									//The host has read the data we provided, let's enter the status phase
									//Free the DMA buffer

									//Let's not be too sure! Check to make sure this is indeed the case!
									if (gActiveEndpointStates[currentEndpoint].maxPacketSize >= 
										(*(PVUS)((UINT)USBD_EPLEN0 + (USBD_EPCFG1 - USBD_EPCFG0) * currentEndpoint)))
										{
											Dprintf ("Entering Status Phase from Data In");
											FreeDMABlock (gActiveEndpointStates[currentEndpoint].dmaBuffer);
											gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
											gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;

											gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_STATUS_PHASE;
											ArmEndpoint (currentEndpoint, EP_OUT);
										}
									break;
								default:
									//ToDo: For non-control endpoints, we need to do our client notification
									//Etc. here
									break;
							} //(gActiveEndpointStates[currentEndpoint].currentState)
						
					} //if (usbd_epintr[currentEndpoint] & USBD_INTR_TC)

				if (usbd_epintr[currentEndpoint] & USBD_INTR_MERR)
					{
						if (gpOnMERR)
							(gpOnMERR)(currentEndpoint);
					}// (usbd_epintr[currentEndpoint] & USBD_INTR_MERR)

				
			} //(currentEndpoint = EP0; currentEndpoint < MAX_PHY_EP; currentEndpoint++)

		//ToDo: Add code to do whatever is necessary to handle other interrupts
	}



/*------------------------------------------------------------
*	HandleNoDataControlRequest
*
*	Parameters:  
*		currentEndpoint - endpoint on which to handle the request.
*
*	Globals Used:
*		gpOnSetupReq, gActiveEndpointStates
*
*	Description:
*		This routine is called to handle a Control Request(SETUP) with
*		no data. If the client has setup a callback, this routine calls
*		the client, with the setup data. And based on the client's reply
*		it either transitions the transfer to status phase, or stalls the
*		endpoint.
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void HandleNoDataControlRequest (UCHAR currentEndpoint)
	{
		//Call the client to let them know!
		if (gpOnSetupReq)
			{
				if ((gpOnSetupReq) (currentEndpoint, &gActiveEndpointStates[currentEndpoint].setupData,
					nil, nil))
					{
						//Success, meaning we are ready to get into the status phase,
						//Which simply involves receiving a zero length DATA OUT packet.
						gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_STATUS_PHASE;
						ArmEndpoint (currentEndpoint, EP_OUT);
					}
				else
					{
						//stall!
						StallEndpoint (currentEndpoint);
						GetControlEndpointReady (currentEndpoint);
					}
			}
		else
			{
				//stall!
				StallEndpoint (currentEndpoint);
				GetControlEndpointReady (currentEndpoint);
			}
	}




/*------------------------------------------------------------
*	HandleDataOutControlRequest
*
*	Parameters:  
*		currentEndpoint - endpoint on which to handle the request.
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		This routine is called to handle a Control Request(SETUP) with
*		"OUT" data. It allocates a buffer based on the wLength provided
*		in the setup request, and arms the endpoint again to recieve the
*		setup data.
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void HandleDataOutControlRequest (UCHAR currentEndpoint)
	{
		//We come here because the host wants to send more data with the control req.
		gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_DATA_OUT_PHASE;

		gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;
		gActiveEndpointStates[currentEndpoint].dmaBuffer = 
			AllocDMABlock (gActiveEndpointStates[currentEndpoint].setupData.wLength);
		if (!gActiveEndpointStates[currentEndpoint].dmaBuffer)
			{
				//Should probably stall!
				StallEndpoint (currentEndpoint);
				GetControlEndpointReady (currentEndpoint);
			}
		else
			{
				//The DMA buffer was allocated
				gActiveEndpointStates[currentEndpoint].dmaBufferSize = gActiveEndpointStates[currentEndpoint].setupData.wLength;
				ArmEndpoint (currentEndpoint, EP_OUT);
			} //(!gActiveEndpointStates[currentEndpoint].dmaBuffer)
}



/*------------------------------------------------------------
*	HandleDataInControlRequest
*
*	Parameters:  
*		currentEndpoint - most of the time this will be 0
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		This routine is called to handle a DATA In requset on the control
*		endpoint. This happens as part of a control transfer requiring
*		data. This routine calls the DispatchDataInControlRequest which
*		will either handle the request itself, or call upon the client
*		to handle the request.
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void HandleDataInControlRequest (UCHAR currentEndpoint)
	{
		//We come here because the host wants to get data with the control req.
		gActiveEndpointStates[currentEndpoint].currentState = EP_STATE_DATA_IN_PHASE;

		gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;
		gActiveEndpointStates[currentEndpoint].dmaBuffer = 
			AllocDMABlock (gActiveEndpointStates[currentEndpoint].setupData.wLength);
		if (!gActiveEndpointStates[currentEndpoint].dmaBuffer)
			{
				//Should probably stall!
				StallEndpoint (currentEndpoint);
				GetControlEndpointReady (currentEndpoint);
			}
		else
			{
				gActiveEndpointStates[currentEndpoint].dmaBufferSize = gActiveEndpointStates[currentEndpoint].setupData.wLength;

				//This could be one of the Get Descriptor requests, let's try to find out, and dispatch
				//accordingl
				if (DispatchDataInControlRequest (currentEndpoint))
					{
						ArmEndpoint (currentEndpoint, EP_IN);
					}
				else
					{
						//Free the DMA buffer
						FreeDMABlock (gActiveEndpointStates[currentEndpoint].dmaBuffer);
						gActiveEndpointStates[currentEndpoint].dmaBuffer = nil;
						gActiveEndpointStates[currentEndpoint].dmaBufferSize = 0;

						//Stall
						StallEndpoint (currentEndpoint);
						GetControlEndpointReady (currentEndpoint);
					}
			} //(!gActiveEndpointStates[currentEndpoint].dmaBuffer)
	}


/*------------------------------------------------------------
*	DispatchDataInControlRequest
*
*	Parameters:
*		endpointNumber - the endpoint on which the request was received
*
*	Globals Used:
*		None
*
*	Description:
*		This routine is called to handle a Control requset
*		with a DATA In phase. By the time this routine is called
*		the DMA buffer is allocated and all setup to be filled
*
*	Returns:
*		true if data is filled in, false if not.
*
*------------------------------------------------------------*/
bool DispatchDataInControlRequest (UCHAR endpointNumber)
	{
		bool sendToClient = true;
		bool retVal = false;
		PUSB_SETUP_DATA pSetupData = &gActiveEndpointStates[endpointNumber].setupData;

		//First see if it is a GET DESCRIPTOR request
		if ((endpointNumber == EP0) && 
			(USB_CONTROL_REQUEST_GET_DIRECTION (pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_IN) &&
			(USB_CONTROL_REQUEST_GET_TYPE (pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_STANDARD) &&
			(USB_CONTROL_REQUEST_GET_RECIPIENT (pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_DEVICE) &&
			(pSetupData->bRequest == USB_GET_DESCRIPTOR)
			)
			{
				switch (HIBYTE(pSetupData->wValue))
					{
						case USB_DEVICE_DESCRIPTOR:
							//Make sure that the index is 0, and the client has registered a callback for
							//get device descriptor
							if ((LOBYTE(pSetupData->wValue) == 0) && (gpOnGetDeviceDescriptor))
								{
									Dprintf ("Get Device Descriptor");
									sendToClient = false;
									retVal = (gpOnGetDeviceDescriptor)(gActiveEndpointStates[endpointNumber].dmaBuffer, 
										&gActiveEndpointStates[endpointNumber].dmaBufferSize);
								}
							break;

						case USB_CONFIGURATION_DESCRIPTOR:
							//Make sure that the index is 0, and the client has registered a callback for 
							//get config descriptor
							Dprintf ("Get Config Descriptor");
							if ((LOBYTE(pSetupData->wValue) == 0) && (gpOnGetConfigDescriptor))
								{
									sendToClient = false;
									retVal = (gpOnGetConfigDescriptor)(gActiveEndpointStates[endpointNumber].dmaBuffer, 
										&gActiveEndpointStates[endpointNumber].dmaBufferSize);
								}
							break;

						case USB_STRING_DESCRIPTOR:
							//Make sure that the the client has registered a callback for 
							//get string descriptor
							Dprintf ("Get String Descriptor");
							if (gpOnGetStringDescriptor)
								{
									sendToClient = false;
									retVal = (gpOnGetStringDescriptor)(LOBYTE(pSetupData->wValue), pSetupData->wIndex, gActiveEndpointStates[endpointNumber].dmaBuffer, 
										&gActiveEndpointStates[endpointNumber].dmaBufferSize);
								}
							break;
					}
			}


		//Everything else will be caught here!
		if (sendToClient)
			{
				if (gpOnSetupReq)
					retVal = ((gpOnSetupReq) (endpointNumber, &gActiveEndpointStates[endpointNumber].setupData,
						gActiveEndpointStates[endpointNumber].dmaBuffer, &gActiveEndpointStates[endpointNumber].dmaBufferSize));
			}

		return(retVal);
		
	}


/*------------------------------------------------------------
*	GetControlEndpointReady
*
*	Parameters:  
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		This routine gets a control endpoint ready to receive the next
*		SETUP request.
*
*	Returns:
*		TRUE on success, FALSE if cannot allocate dma buffer.
*
*------------------------------------------------------------*/
bool GetControlEndpointReady (UCHAR endpointNumber)
	{
		//Move to waiting phase
		gActiveEndpointStates[endpointNumber].currentState = EP_STATE_WAITING_FOR_SETUP;

		//First allocate a buffer
		gActiveEndpointStates[endpointNumber].dmaBufferSize = 0;
		gActiveEndpointStates[endpointNumber].dmaBuffer = AllocDMABlock (gActiveEndpointStates[endpointNumber].defaultMaxTransferSize);

		if (!gActiveEndpointStates[endpointNumber].dmaBuffer)
			return(false);

		gActiveEndpointStates[endpointNumber].dmaBufferSize = gActiveEndpointStates[endpointNumber].defaultMaxTransferSize;

		//Now arm it!
		ArmEndpoint (endpointNumber, EP_OUT);

		return(true);
	}




/*------------------------------------------------------------
*	PrepareINEPForXFer
*
*	Parameters:  
*		endpointNumber - endpoint to prepare
*		ppDataBuffer - pointer to a location to recevied the allocated
*			buffer.
*
*	Globals Used:
*		gActiveEndpointStates
*
*	Description:
*		This routine prepares an endpoint for a transfer by setting the
*		state variables and allocating a block out of the DMA
*		buffer state. The DMA buffer size is determined by
*		the defaultMaxTransferSize setup for the endpoint previously
*		using ActivateEndpoint.
*
*	Returns:
*		true on success, false if the dma buffer could not be allocated.
*
*------------------------------------------------------------*/
bool PrepareINEPForXFer (UCHAR endpointNumber, void **ppDataBuffer)
	{
		//Move to waiting phase
		gActiveEndpointStates[endpointNumber].currentState = EP_STATE_WAITING_FOR_TC;

		//First allocate a buffer
		gActiveEndpointStates[endpointNumber].dmaBufferSize = 0;
		gActiveEndpointStates[endpointNumber].dmaBuffer = AllocDMABlock (gActiveEndpointStates[endpointNumber].defaultMaxTransferSize);

		if (!gActiveEndpointStates[endpointNumber].dmaBuffer)
			return(false);

		gActiveEndpointStates[endpointNumber].dmaBufferSize = gActiveEndpointStates[endpointNumber].defaultMaxTransferSize;
		*ppDataBuffer = gActiveEndpointStates[endpointNumber].dmaBuffer;

		return(true);
	}




/*------------------------------------------------------------
*	StallEndpoint
*
*	Parameters:  
*		endpointToStall - the endpoint to stall
*
*	Globals Used:
*		None
*
*	Description:
*		Stalls the specified endpoint!
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void StallEndpoint (UCHAR endpointToStall)
	{
		USBD_CTRL_REG |= (1 << (endpointToStall + 8));
		asm("ssync;");
	}



//----------------------------------------------------
//
//A very simple heap manager!
//
//



/*------------------------------------------------------------
*	AllocDMABlock
*
*	Parameters:  
*		numBytesToAllocate - number of bytes to allocat.
*
*	Globals Used:
*		gDMAAllocations, gLargestFreeDMABlockOffset, gLargestFreeDMABlock
*
*	Description:
*		Allocates a block out of the DMA buffer space. This routine
*		allocates blocks in increments of 16 bytes.
*
*	Returns:
*		On success address of the allocated block, On failure, nil.
*
*------------------------------------------------------------*/
void *AllocDMABlock (int numBytesToAllocate)
	{
		int bufferIndex, actualBytesToAllocate;
		void *retVal = nil;

		//we need to allocate on a 16 byte boundary!
		if ((ULONG)numBytesToAllocate & 0x000f)
			actualBytesToAllocate = (int)(numBytesToAllocate & 0xfff0) + 16;
		else
			actualBytesToAllocate = numBytesToAllocate;

		/* This is not good as it uses a divide!
		actualBytesToAllocate = (numBytesToAllocate / 16) * 16;
		actualBytesToAllocate += (numBytesToAllocate % 16) ? 16 : 0;
		*/

		if (gTotalDMABuffersAllocated	>= MAX_DMA_BUFFER_ALLOCATIONS)
			return (nil);

		if (gLargestFreeDMABlock < actualBytesToAllocate) 
			return (nil);
		
		//We've got room!
		//Find a slot in our array:
		for (bufferIndex = 0; bufferIndex < MAX_DMA_BUFFER_ALLOCATIONS; bufferIndex++)
			{
				if (gDMAAllocations[bufferIndex].size == 0)
					break;
			}

		gDMAAllocations[bufferIndex].size = actualBytesToAllocate;
		gDMAAllocations[bufferIndex].offset = gLargestFreeDMABlockOffset;
		retVal = &gDMABuffer[gLargestFreeDMABlockOffset];

		gLargestFreeDMABlockOffset += actualBytesToAllocate;
		gLargestFreeDMABlock -= actualBytesToAllocate;

		//Dprintf ("AllocDMABlock: New Free Offset: %x", gLargestFreeDMABlockOffset);

		return (retVal);
	}


/*------------------------------------------------------------
*	FreeDMABlock
*
*	Parameters:  
*		blockToDealloc - address of the block to deallocate.
*
*	Globals Used:
*		gDMAAllocations, gLargestFreeDMABlockOffset, gLargestFreeDMABlock
*
*	Description:
*		This routine deallocates a previously allocated DMA block. Please
*		note that this is currently a very simple heap manager, which
*		treats the DMA buffer more like a stack - assuming that the last
*		allocated block will be deallocated first!
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void FreeDMABlock (void *blockToDealloc)
	{
		UINT offset = (UCHAR *)blockToDealloc - gDMABuffer;
		int numBytesToFree = 0;
		int bufferIndex;

		//We're assuming that the block being deallocated is the last one that was allocated!
		for (bufferIndex = 0; bufferIndex < MAX_DMA_BUFFER_ALLOCATIONS; bufferIndex++)
			{
				if ((gDMAAllocations[bufferIndex].offset == offset) && (gDMAAllocations[bufferIndex].size))
					{
						numBytesToFree = gDMAAllocations[bufferIndex].size;
						gDMAAllocations[bufferIndex].size = 0;

						//We won't need this when we have a proper heap manager
						gLargestFreeDMABlockOffset -= numBytesToFree;
						gLargestFreeDMABlock += numBytesToFree;
						//Dprintf ("FreeDMABlock: New Free Offset: %x", gLargestFreeDMABlockOffset);
						break;
					}
			}

	}



/*------------------------------------------------------------
*	InitDMA
*
*	Parameters:  
*
*	Globals Used:
*		gTotalDMABuffersAllocated, gLargestFreeDMABlockOffset,
*		gLagestFreeDMABlock.
*
*	Description:
*		This routine initializes all the simple DMA block manager!
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void InitDMA (void)
	{
		gTotalDMABuffersAllocated = 0;
		memset (&gDMAAllocations, 0, sizeof (gDMAAllocations));
		gLargestFreeDMABlock = sizeof (gDMABuffer);
		gLargestFreeDMABlockOffset = 0;
	}

/*----------------------------------------------------------------------------------
* $Log: usbdriver.c,v $
* Revision 1.1  2002/10/31 00:30:59  Devendra
* Moved all files in one folder to avoid IDE related problems.
*
* Revision 1.2  2002/10/30 02:36:40  Devendra
* - Added more abstraction to HIDCLASS. So that it's possible to build
*   different devices without making any changes to the HIDCLASS module.
* - Added proper support for handling USB SUSPEND, RESUME, and RESET events.
*
* Revision 1.1  2002/10/22 17:23:39  Devendra
* Rearranged file locations.
*
* Revision 1.5  2002/10/14 01:33:15  Devendra
* Got "custom" device functional with buttons and lights!
*
* Revision 1.4  2002/10/09 17:09:32  Devendra
* - Added support for handling IN endpoints.
* - Added HID descriptors (HID and HID Report)
* - Modified the Interface descriptors to comply to HID Class Spec.
* - Added functionality to use the push-buttons on the Eagle-35 as a mouse (for buttons as well as for movement).
* - The device is now a fully functioning USB Mouse!
*
* Revision 1.3  2002/10/08 20:44:03  Devendra
* - Separated the printf logging to a different c module.
*
* Revision 1.2  2002/10/08 19:57:02  Devendra
* - Added specific callbacks for getting Device, Config, and String descriptors.
* - BugFix: Max. Packet Size was not being set correctly in ArmEndpoint.
* - Added internal buffer based logging to allow real-time "printfs".
* - Added String descriptors.
*
* Revision 1.1  2002/09/20 06:51:30  Devendra
* First Rev.
*
*
*---------------------------------------------------------------------------------*/
